#include <gray.h>
#include <keyboard.h>
#include <string.h>

#include "dungeon.h"
#include "battle.h"
#include "frames.h"
#include "util.h"

//---
//	More declarations...
//---

enum Action
{
	None	= -1,
	Attack	= 0,
	Defend	= 1,
	Skill0	= 2,
	Skill1	= 3,
	Skill2	= 4,
};



//---
//	Static variables.
//
//	This file has a hell lot of thread effects: unrelated functions being
//	called one another without being able to pass any parameter, timers
//	interrupting the flow to update the animation state, frequent drawing
//	demanding constantly-up-to-date variable states, as well as regular
//	interfacing with the dungeon module.
//	The only reasonable ways available to handle such an unpredictable
//	execution flow are:
//
//	- shared variables
//	- a parameter access interface
//
//	A great, well-developed game would implement the latter method, but in
//	our case there is neither the time nor the real necessity to do so. A
//	few variables shared among this module or with the dungeon module
//	should be enough; but I must admit I underestimated the amount of data
//	requiring treatment at the same time.
//
//	We end up with quite a few variables there, though this is not a
//	problem unless they are not handled correctly (which is not the case
//	here, of course ;) ).
//---

// Position of cursors. sel_action is the action bar cursor (keys F1-F5), while
// sel_target is the id (between 0 and 2) of the enemy currently selected.
static enum Action sel_action = None;
static int sel_target = 0;

// Current player and visual effect animation frames.
extern struct AnimationFrame * volatile frame_player;
static struct AnimationFrame * volatile frame_visual = NULL;

// Entities fighting (BEs stands for Battle Entities).
static struct BattleEntity *bes[4]	= { NULL, NULL, NULL, NULL };
// Some shorthands will make code easier to read.
static struct BattleEntity *player;
static struct BattleEntity **enemies = bes + 1;

// Some animation state info.
// Entity currently attacking (counterpart for the frame_player animation,
// except that this animation is minimal), that also indicates that the player
// should move to its target).
static volatile int entity_attacking = -1;
// Target of the current visual effect (0-3).
static volatile int visual_target = -1;
// Time delay to pause level-2 animations for a short amount of time, 
static volatile int time_delay = 0;
// Monsters coming from the right side of the screen at the beginning.
static volatile int opening_move = 0;

// This is for drawing. The draw() function is indirectly called by the timer,
// which cannot take any argument.
extern Image *img_background;

// Current battle context. This basically allows tutorial text showing up
// amidst of a battle. At each step is associated a NULL-terminated string
// array, which is displayed through the dungeon.msg() function when the
// conditions are fulfilled. A battle context item may only be used once.
static BattleContext *context = NULL;

// Message box parameters.
extern volatile int mbox;
extern volatile int mbox_y;
extern volatile reference_t mbox_msg[2];



//---
//	Useful functions.
//---

/*
	doContext()
	Checks if the current context has a string available and uses it. This
	function must be called with BattleStep_Initialize as argument to
	initialize the record.
*/
static void doContext(enum BattleStep step)
{
	static uint8_t done[BattleStep_StepCount];

	if(step == BattleStep_Initialize)
	{
		memset(done, 0, BattleStep_StepCount);
		return;
	}
	if(!context) return;
	if(step < 0 || step >= BattleStep_StepCount) return;
	if(done[step]) return;

	shared_messages((*context)[step]);
	done[step] = 1;
}

/*
	animsRunning()
	Returns non-zero if there are still animations running. The higher
	value returned, the more blocking and volatile animations. For
	instance, animations of level 2 are battle animations, but animations
	of level 1 are background animations.
*/
static int animsRunning(void)
{
	int i;

	// Opening animation.
	if(opening_move) return 3;

	// Battle animations.
	if(time_delay || entity_attacking > 0 || frame_visual) return 2;
	// Life bar animations.
	for(i = 0; i < 4; i++) if(bes[i] && bes[i]->disp_hp != bes[i]->hp)
		return 2;
	// Player mana bar.
	if(player->disp_mp != player->mp) return 2;

	// Player two-frame idle animation (or any other...).
	if(frame_player) return 2 - frames_isIdle(frame_player);

	// Everything went correctly, nothing left to handle! \o/
	return 0;
}

/*
	wait()
	Waits for current animations to end. The threshold indicate which kind
	of animations we are waiting for.
*/
static void wait(int threshold)
{
	while(animsRunning() >= threshold) sleep();
}

/*
	isAlive()
	Checks if the given entity is alive.
*/
static int isAlive(int no)
{
	return (no >= 0 && no <= 3 && bes[no] && bes[no]->hp > 0);
}

/*
	enemyAlive()
	Returns the id of an enemy, which is alive, or 0 if only the player
	is left...
*/
static int enemyAlive(void)
{
	for(int i = 1; i <= 3; i++) if(isAlive(i)) return i;
	return 0;
}

/*
	enemiesAlive()
	Returns the number of enemies alive.
	"All this *slurp* 'divine' powers were *slurp* too ugly for us"
	-- a defeated slime
*/
static int enemiesAlive(void)
{
	return isAlive(1) + isAlive(2) + isAlive(3);
}



//---
//	Drawing.
//---

/*
	draw_position()
	Returns the supposed position of the given entity (0-3), taking in
	consideration the number of monsters and the animations.
*/
static void draw_position(int entity, int *x, int *y)
{
	int enemy_pos[][2] = {
		{   88, 43  },
		{  111, 34  },
		{  111, 54  },
		{   95, 43  },
	};
	// Number of enemies.
	int count = 3 - !enemies[1] - !enemies[2];
	int row;

	// Attacking player.
	if(!entity && !entity_attacking)
	{
		draw_position(sel_target, x, y);
		*x -= 20;
		return;
	}
	// Idle player.
	else if(!entity)
	{
		if(x) *x = DUNGEON_PLAYER_X;
		if(y) *y = DUNGEON_PLAYER_Y;
		return;
	}

	// The row refers to the coordinate pair from the enemy_pos array,
	// which will be used. These conditions take note of the monster count
	// to display monsters at best positions.
	if(count == 1) row = 3;			// Alone in the middle
	if(count == 2) row = entity;		// Next to each other
	if(count == 3) row = entity - 1;	// Triangle configuration

	if(x) *x = enemy_pos[row][0] + opening_move;
	if(y) *y = enemy_pos[row][1];
}

/*
	draw_monster()
	Draws monster number 'no' out of 'count' monsters. This function takes
	account of the monster number to use all the available space with
	different configurations. Also draws a life bar and a selection
	indicator if the monster is selected.
*/
static void draw_monster(struct BattleEntity *e, int no, int count)
{
	extern Image res_cursor2_start;
	struct Sprite *s;
	int x, y;

	if(!e || no > count) return;
	if(e->disp_hp <= 0) return;
	s = e->sprite;

	draw_position(no, &x, &y);
	x -= s->dx;
	y -= s->dy;

	// The enemy-attacking effect is a bit curious. It consists in
	// displaying the same sprite three times at slightly different
	// positions, which creates a sort of shadow.
	if(no == entity_attacking)
	{
		gimage_part(x - 1, y, s->image, s->left, s->top, s->width,
			s->height);
		gimage_part(x + 1, y, s->image, s->left, s->top, s->width,
			s->height);
	}
	if(e->disp_hp > 0)
		gimage_part(x, y, s->image, s->left, s->top, s->width,
			s->height);

	// Enemy life bar.
	gline(x, y - 3, x + s->width - 1, y - 3, Color_Light);
	gline(x, y - 3, x + e->disp_hp * (s->width - 1) / e->max_hp, y - 3,
		Color_Black);

	// Cursor, if the enemy is selected.
	if(sel_action != None && sel_target == no)
		gimage(x - 10, y + s->dy - 1, &res_cursor2_start);
}

/*
	battle_draw()
	Draws the whole interface according to the timed parameters.
*/
void battle_draw(void)
{
	extern Image
		res_actions_start,
		res_cursor_start,
		res_player_health_start;
	struct Sprite *s;

	int i, count = 0, x, y;
	while(count < 3 && enemies[count]) count++;

	// Background.
	gclear();
	gimage(0, 0, img_background);

	// Player sprite and monsters.
	if(frame_player)
	{
		draw_position(0, &x, &y);
		s = frame_player->sprite;
		gimage_part(x - s->dx, y - s->dy, s->image, s->left, s->top,
			s->width, s->height);
	}
	for(i = 0; i < count; i++) draw_monster(enemies[i], i + 1, count);

	// Visual effect, if running.
	if(frame_visual)
	{
		draw_position(visual_target, &x, &y);
		s = frame_visual->sprite;
		gimage_part(x - s->dx, y - s->dy, s->image, s->left, s->top,
			s->width, s->height);
	}

	// Action bar at the bottom. We assume that the player's skills are
	// always in the order: fire, cure, special.
	int skill_count = (player->skills[0] != Skill_None)
		+ (player->skills[1] != Skill_None)
		+ (player->skills[2] != Skill_None);
	gimage_part(3, 55, &res_actions_start, 0, 0, 21 + 12 * skill_count, 9);
	if(sel_action != None)
		gimage(1 + 12 * sel_action, 54, &res_cursor_start);

	// Player life and mana bars.
	gimage(1, 1, &res_player_health_start);

	#define d Color_Dark
	if(player->disp_hp > 0)
	{
		gline(12, 3, 12 + (player->disp_hp*35 / player->max_hp), 3, d);
		gline(12, 4, 12 + (player->disp_hp*35 / player->max_hp), 4, d);
	}
	if(player->disp_mp > 0)
		gline(11, 6, 11 + (player->disp_mp*34 / player->max_mp), 6, d);
	#undef d

	// Enemy name and level.
	if(sel_action != None && sel_target)
	{
		gclear_area(54, 0, 127, 7);
		gprint(55, 2, "LV.%d %s", bes[sel_target]->level,
			bes[sel_target]->name);
	}
}



//---
//	Timer management.
//---

/*
	battle_timer()
	Forwards time by a step: animations and sprites change here.
*/
void battle_timer(void)
{
	// These variables count occurrence of certain events and perform the
	// timeout check. This allows animations to have various durations.
	// Let's limit the scope as much as possible: these variables do not
	// need to be accessible in a larger scope than inside this function.

	static int past_player = 0;	// Frames for frame_player
	static int past_visual = 0;	// Frames for frame_visual
	static int past_attack = 0;	// Frames for entity_attacking

	// These variables track changes between the last frame variables and
	// the current ones, when this is needed. This allows efficient
	// modification of player animations in other parts of the code.

	static struct AnimationFrame *local_player = NULL;
	static struct AnimationFrame *local_visual = NULL;
	static int local_attacking = -1;

	struct BattleEntity *e;
	int redraw = 0, i;

	// Time delay, used to separate level-2 battle animations.
	if(time_delay > 0) time_delay--;
	// Opening move: monsters come from the right of the screen.
	if(opening_move > 0) opening_move--, redraw = 1;
	// Message box: tutorials and the like.
	if(mbox && mbox_y != DUNGEON_MBOX) mbox_y++, redraw = 1;
	if(!mbox && mbox_y) mbox_y--, redraw = 1;

	// Checking player animation timeout.
	if(frame_player != local_player)
	{
		past_player = 0;
		local_player = frame_player;
		redraw = 1;
	}
	else if(frame_player) if(++past_player >= frame_player->duration)
	{
		past_player = 0;
		frame_player = frame_player->next;
		redraw = 1;
	}

	// Checking visual effect timeout.
	if(frame_visual != local_visual)
	{
		past_visual = 0;
		local_visual = frame_visual;
		redraw = 1;
	}
	else if(frame_visual) if(++past_visual >= frame_visual->duration)
	{
		past_visual = 0;
		frame_visual = frame_visual->next;
		redraw = 1;
	}

	// Checking life bar animations.
	for(i = 0; i < 3; i++) if((e = enemies[i]))
	{
		// This would mean the life bar is still being animated.
		if(e->disp_hp != e->hp)
		{
			e->disp_hp += (e->disp_hp > e->hp) ? (-1) : (1);
			redraw = 1;
		}
	}

	// Checking player's life and mana bars.
	if(player->disp_hp != player->hp)
	{
		player->disp_hp += (player->disp_hp > player->hp) ? (-1) : (1);
		redraw = 1;
	}
	if(player->disp_mp != player->mp)
	{
		player->disp_mp += (player->disp_mp > player->mp) ? (-1) : (1);
		redraw = 1;
	}

	// Checking enemies currently attacking.
	if(entity_attacking >= 0)
	{
		// Re-draw interface the first time.
		if(entity_attacking != local_attacking)
		{
			past_attack = 0;
			redraw = 1;
		}
		local_attacking = entity_attacking;
		// We can reset entity_attacking before drawing because the
		// animation lasts longer than a one frame. We do not do this
		// for the player because its position shift must last during
		// a full animation time, which is not known. The function that
		// sets entity_attacking to 0 will have to reset it to -1 after
		// the animation is finished.
		past_attack++;
		if(entity_attacking != 0 && past_attack >= 6)
		{
			entity_attacking = -1;
			past_attack = 0;
		}
	}

	// You were expecting it, weren't you?
	if(redraw) shared_draw();
}



//---
//	Actions.
//---

/*
	do_action()
	Executes a command validated by SHIFT or by pressing two times the same
	F-key (except for 'Defend').
*/
static void do_action(void)
{
	enum Action action = sel_action;
	int max, i, j, n;
	int be_count = 4 - !enemies[2] - !enemies[1];

	// Order of fighters in the turn. Fast entities are found early when
	// reading the array.
	int fighters[4] = { 0, 1, 2, 3 };
	// Action for every entity.
	struct
	{
		int skill;	// 0: physical attack, other values: skills
		int target;
	} actions[4] = { { 0, 0 }, { 0, 0 }, { 0, 0 }, { 0, 0 } };

	//---
	//	Preparing turn.
	//---

	if(action == Defend)
	{
		player->stance = Stance_Guard;
		actions[0].skill = 0;
		actions[0].target = 0;
	}
	else
	{
		player->stance = Stance_Fight;
		actions[0].skill = (action == Attack) ? (0) : (action - 1);
		actions[0].target = sel_target;
	}

	//---
	//	Determining what the enemies will do.
	//---

	for(i = 1; bes[i] && i <= 3; i++)
	{
		/* Somehow this is the AI */
		bes[i]->stance = Stance_Fight;
		actions[i].target = 0;

		// Use a skill when possible.
		if(bes[i]->skills[0] != Skill_None &&
			bes[i]->mp >= skill_cost(bes[i]->skills[0]))
		{	
			actions[i].skill = 1;
		}
		// Otherwise just attack.
		else actions[i].skill = 0;
	}

	//---
	//	Computing the turn order.
	//---

	// Sorting the turn order array according to the speed stat. Entities
	// will act in the order specified by this array for the given turn.
	for(i = 0; i < be_count; i++)
	{
		// Entity number (0-3) whose speed is the best among the
		// remaining ones.
		max = i;

		// Finding the quickest entity.
		for(j = i + 1; j < be_count; j++)
		{
			if(bes[fighters[j]]->spe > bes[fighters[max]]->spe)
				max = j;
		}

		// Swapping entity no.i with entity no. max.
		int tmp = fighters[i];
		fighters[i] = fighters[max];
		fighters[max] = tmp;
	}

	//---
	//	Fight!
	//---

	// Hiding the cursors before doing anything else.
	sel_action = None;

	for(i = 0; i < be_count; i++)
	{
		n = fighters[i];
		// This should not happen but still...
		if(!bes[n]) continue;

		// Entities who defend basically do nothing...
		if(bes[n]->stance == Stance_Guard) continue;
		// Battle entities right, paragraph 4.2.b: No one shall attempt
		// to target a dead enemy.
		if(!isAlive(actions[n].target)) continue;
		// Law of nature, art. 0: No one may rise from the dead.
		if(!isAlive(n)) continue;

		if(!actions[n].skill)
		{
			// Attack animation for monsters.
			if(n)
			{
				entity_attacking = n;
				wait(2);
			}
			// Attack animation for the player.
			else
			{
				// The player suddenly disappears!
				frame_player = NULL;
				visual_target = 0;
				frame_visual = frames_visual(Visual_Disappear);
				wait(2);

				// He re-appears and strikes!
				entity_attacking = 0;
				frame_player = frames_anim(Animation_Sword);
				visual_target = actions[n].target;
				frame_visual = frames_visual(Visual_Sword);
				wait(2);

				entity_attacking = -1;
			}

			// Damage animation.
			visual_target = actions[n].target;
			frame_visual = frames_visual(Visual_Damage);
			wait(2);

			entity_hit(bes[n], bes[actions[n].target]);
		}
		else
		{
			int skill = bes[n]->skills[actions[n].skill - 1];
			bes[n]->mp -= skill_cost(skill);

			// Attack animation for monsters.
			if(n)
			{
				entity_attacking = n;
				wait(2);
			}
			else
			{
				frame_player = frames_anim(Animation_Skill);
				wait(2);
			}

			// Visual effect.
			visual_target = actions[n].target;
			frame_visual = frames_visual(skill_visual(skill));
			wait(2);

			entity_skill(bes[actions[n].target], skill);
			if(!n) frame_player = frames_anim(Animation_Idle);
		}

		// Life bar animations.
		wait(2);

		// Possible death animations.
		if(actions[n].target && !isAlive(actions[n].target))
		{
			visual_target = actions[n].target;
			frame_visual = frames_visual(Visual_Death);
			wait(2);
		}

		if(!n && player->stance == Stance_Fight)
		{
			doContext(BattleStep_HitLanded);
			if(!isAlive(actions[n].target))
				doContext(BattleStep_EnemyKilled);
		}

		// Stop the turn if the battle is finished.
		if(!isAlive(0) || !enemiesAlive()) break;

		time_delay = 24;
		wait(2);
	}

	doContext(BattleStep_TurnOver);
}


//---
//	The game's core.
//---

/*
	battle()
	Fight!
*/
int battle(BE *arg_player, BE *enemy1, BE *enemy2, BE *enemy3,
	 BattleContext *arg_context)
{
	extern Font res_font_arena_start;

	int key = 0, redraw = 0;
	int count = 3 - !enemy3 - !enemy2; // number of enemies
	int n;

	if(!arg_player || !enemy1) return 0;
	text_configure(&res_font_arena_start, Color_Black);

	//---
	//	Initializing the static variables.
	//---

	sel_action = None;
	sel_target = 0;

	frame_visual = NULL;

	bes[0] = arg_player;
	player = arg_player;
	enemies[0] = enemy1;
	enemies[1] = enemy2;
	enemies[2] = enemy3;

	entity_attacking = -1;
	visual_target = -1;
	opening_move = 12;

	context = arg_context;

	doContext(BattleStep_Initialize);

	//---
	//	Let's do it!
	//---

	shared_draw();
	wait(3);
	doContext(BattleStep_Opened);

	while(isAlive(0) && enemiesAlive())
	{
		if(redraw) shared_draw();

		wait(2);

		key = getkey_opt(Getkey_NoOption, 0);
		if(key == KEY_MENU) return 1;
		redraw = 1;

		switch(key)
		{
		// Handling the target cursor over the monster triangle is
		// quite a pain if you want to code it properly. For now, let's
		// do something that works. Implementing a wider-range method
		// would be useless and heavy...
		case KEY_LEFT:
			if(sel_target > 1 && count == 3 && isAlive(1))
				sel_target = 1;
			else redraw = 0;
			break;
		case KEY_RIGHT:
			if(sel_target == 1 && count == 3)
			{
				if(isAlive(2)) sel_target = 2;
				else if(isAlive(3)) sel_target = 3;
				else redraw = 0;
			}
			else redraw = 0;
			break;
		case KEY_DOWN:
			if(count >= 2)
			{
				if(sel_target == count - 1 && isAlive(count))
					sel_target = count;
				else if(sel_target == 1 && isAlive(3))
					sel_target = 3;
			}
			else redraw = 0;
			break;
		case KEY_UP:
			if(count >= 2)
			{
				if(sel_target == count && isAlive(count - 1))
					sel_target = count - 1;
				else if(sel_target == 1 && isAlive(2))
					sel_target = 2;
			}
			else redraw = 0;
			break;

		// Enough talk, time to fight! We can do the checks in that
		// order because skills are unlocked from left to right.
		case KEY_F5:
		case KEY_F3:
		case KEY_F1:
			n = 6 - (key >> 4);

			if(n >= 2 && n < 5) if(player->skills[n - 2] ==
				Skill_None || player->mp < skill_cost(
				player->skills[n - 2])) break;

			// When pressing two times the same key, validate the
			// command.
			if(sel_action == 6 - (key >> 4)) do_action();

			// Otherwise, and if there's more than one enemy alive,
			// display a cursor to choose a target.
			else
			{
				sel_action = 6 - (key >> 4);
				if(!sel_target || !isAlive(sel_target))
					sel_target = enemyAlive();

				/* Un-comment this to automatically target the
				lone enemy. */
				// if(enemiesAlive() == 1) do_action();
			}
			break;

		// The 'Defend' command needn't any target selection, so
		// perform it immediately.
		case KEY_F2:
			sel_action = Defend;
			do_action();
			break;

		// The 'Heal' skill doesn't need any target either.
		case KEY_F4:
			if(player->skills[1] == Skill_None || player->mp <
				skill_cost(player->skills[1])) break;
			sel_action = Skill1;
			sel_target = 0;
			do_action();
			break;

		// Cancel selection.
		case KEY_ALPHA:
			sel_action = None;
			sel_target = 0;
			break;

		// Please press a key that does something!
		default:
			redraw = 0;
			break;
		}
	}

	//---
	//	Good riddance. Now let's dance a bit and un-initialize things.
	//	Oh wait, maybe I'm dead actually...
	//---

	wait(2);

	if(isAlive(0))
	{
		frame_player = frames_anim(Animation_Victory);
		wait(2);
		frame_player = frames_anim(Animation_Idle);
	}
	else
	{
		frame_player = frames_anim(Animation_Dead);
		time_delay = 48;
		wait(2);
	}

	doContext(isAlive(0) ? BattleStep_Victory : BattleStep_Dead);
	return 0;
}
