#include <timer.h>
#include <gray.h>
#include <keyboard.h>
#include <stdarg.h>
#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <mpu.h>

#include "dungeon.h"
#include "frames.h"
#include "save.h"

#define	min(a, b) ((a) < (b) ? (a) : (b))

// Forward declarations: because of the order of functions in this file, some
// prototyping is required.
static void dungeon_draw(void);
static void dungeon_timer(void);



//---
//	Static variables.
//	Always the thread effects. The battle file has the worst but still,
//	some are needed there.
//---

// Message box parameters.
volatile int mbox = 0;
volatile int mbox_y = 0;
volatile reference_t mbox_msg[2] = { NULLREF, NULLREF };

// Are we in battle or waiting for enemies to rush?
static volatile int battle_running = 0;

// Dungeon background and player animation frame.
Image *img_background = NULL;
struct AnimationFrame * volatile frame_player = NULL;



//---
//	Shared methods (called by timers and generics).
//---

void shared_draw(void)
{
	char buffer[DUNGEON_MBOX_CPL + 1];

	if(battle_running) battle_draw();
	else dungeon_draw();

	// Message box, drawn for both modules.
	if(mbox_y)
	{
		gline(0, 64 - mbox_y, 127, 64 - mbox_y, Color_Black);
		gclear_area(0, 65 - mbox_y, 127, 63);
	}
	if(mbox_msg[0].begin)
	{
		memset(buffer, 0, DUNGEON_MBOX_CPL + 1);
		refcpy(buffer, mbox_msg[0]);
		gtext(2, 63 - mbox_y + 4, buffer);
	}
	if(mbox_msg[1].begin)
	{
		memset(buffer, 0, DUNGEON_MBOX_CPL + 1);
		refcpy(buffer, mbox_msg[1]);
		gtext(2, 63 - mbox_y + 11, buffer);
	}

	gupdate();
}

void shared_timer(void)
{
	if(battle_running) battle_timer();
	else dungeon_timer();
}

/*
	msg_line()
	Extracts the first line of text, which is available from *string and
	returns it in a reference. Does not return more than DUNGEON_MBOX_CPL
	characters and only splits on whitespaces. Always take at least one
	character.
	Also updates *string so that it points after the portion of text, which
	has been returned.
*/
static reference_t msg_line(const char **string)
{
	reference_t ref = { *string, *string };
	const char *ptr = *string + 1;

	if(!**string) return NULLREF;

	while(ptr - *string < DUNGEON_MBOX_CPL && *ptr)
	{
		if(isspace(*ptr))
		{
			ref.end = ptr;
			while(*ptr && isspace(*ptr)) ptr++;
		}
		else ptr++;
	}
	if(!*ptr) ref.end = ptr;

	*string = ref.end;
	while(**string && isspace(**string)) (*string)++;
	return ref;
}

/*
	shared_messages()
	Execution-blockingly (is it even correct to use this word?) displays
	messages.
*/
int shared_messages(const char **messages)
{
	const char *sentence;
	int key = KEY_NONE;

	if(!messages) return 0;
	if(!messages[0]) return 0;
	sentence = *messages++;

	// Make the message box appear.
	mbox = 1;
	do sleep();
	while(mbox_y != DUNGEON_MBOX);

	while(sentence && *sentence)
	{
		mbox_msg[0] = msg_line(&sentence);
		mbox_msg[1] = msg_line(&sentence);

		shared_draw();
		do key = getkey_opt(Getkey_NoOption, 0);
		while(key != KEY_ALPHA && key != KEY_SHIFT && key != KEY_EXE
			&& key != KEY_MENU);
		if(key == KEY_MENU) break;

		// Get another sentence once the current one has been
		// completely displayed.
		if(!*sentence) sentence = *messages++;
	}

	mbox = 0;
	mbox_msg[0] = NULLREF;
	mbox_msg[1] = NULLREF;

	if(key == KEY_MENU)
	{
		mbox_y = 0;
		return 1;
	}
	else
	{
		do sleep();
		while(mbox_y);
		return 0;
	}
}



//---
//	Local drawing.
//---

static void dungeon_draw(void)
{
	struct Sprite *s;

	gclear();
	gimage(0, 0, img_background);

	if(frame_player)
	{
		s = frame_player->sprite;
		gimage_part(DUNGEON_PLAYER_X - s->dx, DUNGEON_PLAYER_Y - s->dy,
			s->image, s->left, s->top, s->width, s->height);
	}
}



//---
//	Local timer event handling.
//---

static void dungeon_timer(void)
{
	static int past_player = 0;
	int redraw = 0;

	if(mbox && mbox_y != DUNGEON_MBOX) mbox_y++, redraw = 1;
	if(!mbox && mbox_y) mbox_y--, redraw = 1;

	if(frame_player) if(++past_player >= frame_player->duration)
	{
		past_player = 0;
		frame_player = frame_player->next;
		redraw = 1;
	}

	if(redraw) shared_draw();
}



//---
//	Dungeon levels.
//---

/*
	fight()
	"I'm gonna pulverize them and smash them down!" -- Carrus
	Returns non-zero if Carrus dies in battle, or the user leaves the
	battle.
*/
static int fight(BattleContext *context, BE *player,
	enum Character c1, int l1,
	enum Character c2, int l2,
	enum Character c3, int l3)
{
	struct BattleEntity enemies[3];
	int x;

	enemies[0] = entity(c1, l1);
	enemies[1] = entity(c2, l2);
	enemies[2] = entity(c3, l3);

	battle_running = 1;
	x = battle(player,
		c1 != Character_None ? &enemies[0] : NULL,
		c2 != Character_None ? &enemies[1] : NULL,
		c3 != Character_None ? &enemies[2] : NULL,
		context
	);
	battle_running = 0;

	return (x || player->hp <= 0);
}

int dungeon_tutorial(void)
{
	BattleContext context0 = {
		// Battle opened.
		{
			"YOU KNOW WELL HOW TO FIGHT ENEMIES ALREADY, BUT "
			"LET'S REVIEW THE BASICS.",
			"PRESS F1 TO ATTACK, THEN SELECT YOUR TARGET AND "
			"PRESS F1 AGAIN.",
			NULL
		},
		// Hit landed.
		{
			"THAT'S GOOD, BUT WATCH OUT, FOR ENEMIES WILL ATTACK "
			"YOU AS WELL.",
			NULL
		},
		// Turn over.
		{
			"KEEP AN EYE ON YOUR HP BAR, OR YOU WILL DIE SOON.",
			"PRESS F2 TO DEFEND AGAINST ENEMIES.",
			NULL
		},
		// Enemy killed.
		{ NULL },
		// Victory.
		{
			"THAT WAS A PRETTY GOOD START.",
			"LET'S SEE HOW YOU DO AGAINST STRONGER FOES.",
			NULL
		},
		// Death.
		{ NULL },
	};
	BattleContext context1 = {
		// Battle opened.
		{
			"TAKE CARE, THIS NEW ENEMY KNOWS A THUNDER SPELL.",
			NULL
		},
		// Hit landed.
		{ NULL },
		// Turn over.
		{
			"YOU CAN USE SKILLS, TOO, USING F3-F5. TRY THIS FIRE "
			"SPELL FOR INSTANCE.",
			NULL
		},
		// Enemy killed.
		{ NULL },
		// Victory
		{
			"YOU ALSO HAVE A CURE SPELL AND A SPECIAL ABILITY.",
			"ALL OF THIS COST MP TO USE, AND YOU CANNOT RESTORE "
			"YOUR MP UNTIL YOU FINISH A LEVEL, SO USE THEM "
			"WISELY.",
			"YOU'LL GET STRONGER EACH TIME YOU FINISH A LEVEL. "
			"NOW GO!",
			NULL
		},
		// Death.
		{
			"THOSE ENEMIES ARE NOT VERY STRONG. REMEMBER TO USE "
			"YOUR FIRE SPELL TO DESTROY THE MIDDLE ROBOT.",
			NULL
		},
	};
			

	const char *msg_init[] = {
		"YOU ARE CARRUS, A SKILLED FIGHTER CAPTURED BY ROME.",
		"YOU HAVE THE BLESSING OF THE GODS. JUPITER HIMSELF HAS LENT "
		"YOU GREAT POWER IN THE PAST.",
		"YOUR MISSION IS TO WIELD YOUR DIVINE POWERS TO ESCAPE THE "
		"CITY.",
		NULL
	};

	BE player = entity(Character_Player, -1);

	if(shared_messages(msg_init)) return 0;

	if(fight(&context0, &player,
		Character_Robot1, 1,
		Character_Robot1, 1,
		Character_None,	  0)) return 0;

	player.skills[0] = Skill_Fire;

	if(fight(&context1, &player,
		Character_Robot2, 1,
		Character_Robot1, 2,
		Character_Robot1, 2)) return 0;

	// Rewards time!
	save()->entity.skills[0] = Skill_Fire;
	save()->entity.skills[1] = Skill_Cure;
	save()->entity.skills[2] = Skill_Special;

	save()->entity.max_hp	= 120;
	save()->entity.max_mp	= 35;
	save()->entity.atk	= 16;
	save()->entity.def	= 6;

	if(save()->dungeons_conquered < 1) save()->dungeons_conquered = 1;
	return 0;
}

int dungeon_prison(void)
{
	BattleContext default_context = {
		// Battle opened.
		{ NULL },
		// Hit landed.
		{ NULL },
		// Turn over.
		{ NULL },
		// Enemy killed.
		{ NULL },
		// Victory.
		{ NULL },
		// Death.
		{
			"YOU NEED TO SAVE UP SOME MP FOR THE SOLDIERS.",
			"THE SPECIAL ABILITY IS A POWERFUL PHYSICAL ATTACK, "
			"BUT MAGIC IGNORES ARMOR.",
			NULL
		},
	};
	const char *msg_init[] = {
		"THE PRISON IS FULL OF DANGEROUS ANIMALS. DO NOT USE ALL YOUR "
		"MP IN THE FIRST BATTLES!",
		NULL
	};

	const char *msg_guards[] = {
		"RIGHT, YOU'RE ALMOST THERE! THERE ARE ONLY SOME GUARDS.",
		NULL
	};

	const char *msg_done[] = {
		"GOOD. NOW YOU NEED TO ESCAPE FROM THE COLOSSEUM. THIS WON'T "
		"BE EASY!",
		NULL
	};

	BE player = entity(Character_Player, -1);

	if(shared_messages(msg_init)) return 0;

	if(fight(&default_context, &player,
		Character_Slime,  1,
		Character_Slime,  1,
		Character_Slime,  1)) return 0;

	if(fight(&default_context, &player,
		Character_Slime,   2,
		Character_Spider,  1,
		Character_Spider,  1)) return 0;

	if(fight(&default_context, &player,
		Character_Spider,  2,
		Character_Rat,     1,
		Character_Slime,   2)) return 0;

	if(shared_messages(msg_guards)) return 0;

	if(fight(&default_context, &player,
		Character_Soldier, 1,
		Character_Soldier, 1,
		Character_None,    0)) return 0;

	shared_messages(msg_done);

	// Rewards time!
	save()->entity.max_hp	= 250;
	save()->entity.max_mp	= 60;
	save()->entity.atk	= 24;
	save()->entity.def	= 10;

	if(save()->dungeons_conquered < 2) save()->dungeons_conquered = 2;
	return 0;
}

int dungeon_colosseum(void)
{
	BattleContext default_context = {
		// Battle opened.
		{ NULL },
		// Hit landed.
		{ NULL },
		// Turn over.
		{ NULL },
		// Enemy killed.
		{ NULL },
		// Victory.
		{ NULL },
		// Death.
		{
			"THERE'S NOT MUCH YOU CAN DO HERE EXCEPT CAREFULLY "
			"MANAGE YOUR LIFE AND MANA.",
			NULL
		},
	};

	BattleContext context0 = {
		// Battle opened.
		{
			"REA... REALLY ? o_O",
			NULL,
		},
		// Hit landed.
		{ NULL },
		// Turn over.
		{
			"THAT'S NO GOOD. YOU MUST STOP IT QUICKLY!",
			NULL,
		},
		// Enemy killed.
		{ NULL },
		// Victory.
		{
			"WELL DONE! YOU REALLY ARE THE WARRIOR JUPITER TOLD.",
			NULL,
		},
		// Death.
		{
			"THIS LION IS A TOUGH OPPENENT. I RECOMMEND YOU "
			"PREPARE CAREFULLY AND SAVE SOME FIRE MAGIC.",
			NULL
		},
	};

	const char *msg_init[] = {
		"BE CARFUL, THERE WILL BE MANY SOLDIERS AROUND THE COLOSSEUM.",
		NULL
	};

	const char *msg_done[] = {
		"A MERCHANT WILL TAKE YOU OUT OF THE EMPIRE, SO THAT YOU "
		"JOIN FOREIGN FORCES.",
		"FREEDOM AWAITS YOU, UNDER THE MIGHTY STARS ABOVE.",
		NULL
	};

	BE player = entity(Character_Player, -1);

	if(shared_messages(msg_init)) return 0;

	if(fight(&default_context, &player,
		Character_Soldier, 1,
		Character_Soldier, 1,
		Character_None,    0)) return 0;

	if(fight(&default_context, &player,
		Character_Soldier, 2,
		Character_Soldier, 2,
		Character_None,    0)) return 0;

	if(fight(&default_context, &player,
		Character_Guard,   2,
		Character_Soldier, 1,
		Character_Soldier, 1)) return 0;

	if(fight(&context0, &player,
		Character_Lion,    1,
		Character_None,    0,
		Character_None,    0)) return 0;

	shared_messages(msg_done);

	if(save()->dungeons_conquered < 3) save()->dungeons_conquered = 3;
	return 1;
}

int (*dungeons[3])(void) = {
	dungeon_tutorial,
	dungeon_prison,
	dungeon_colosseum,
};



//---
//	Dungeon initialization.
//---

/*
	dungeon()
	"Don't tell me Jupiter hadn't foreseen this would be a piece of cake"
	-- Carrus
*/
int dungeon(int dungeon_no)
{
	extern Image
		res_wall0_start,
		res_wall1_start,
		res_wall2_start;
	Image *walls[] = {
		&res_wall0_start,
		&res_wall1_start,
		&res_wall2_start
	};
	int show_credits;

	// Initializing static variables.
	mbox = 0;
	mbox_y = 0;
	img_background = walls[dungeon_no];

	frame_player = frames_anim(Animation_Idle);

	// Initializing environment...
	screen_setContrast(isSH3() ? 160 : 162);
	gray_start();
	timer_start(TIMER_USER, DUNGEON_DELAY, TIMER_Po_256, shared_timer, 0);

	shared_draw();
	show_credits = (*dungeons[dungeon_no])();

	timer_stop(TIMER_USER);
	gray_stop();

	return show_credits;
}
