#include <display.h>
#include <gray.h>
#include <string.h>

#include "util.h"

/*
	refcpy()
	Copies a reference to a string.
*/
char *refcpy(char *str, reference_t ref)
{
	int length = ref.end - ref.begin;
	strncpy(str, ref.begin, length);
	str[length] = 0;

	return str;
}

// Yeah, we actually can't read the contrast from the screen registers (this
// issue still needs investigation but was known to occur within MonochromeLib)
// so we'll just have to use a cached value.
static int contrast;

/*
	sleep()
	Sleeps until an interrupt is triggered.
*/
void sleep(void)
{
	__asm__("sleep\n\t"); // bet you weren't expecting this one
}

/*
	screen_setContrast()
	Changes the screen contrast setting.
*/
void screen_setContrast(int new_contrast)
{
	volatile char *selector	= (char *)0xb4000000;
	volatile char *data	= (char *)0xb4010000;

	contrast = new_contrast;

	*selector = 6;
	*data = new_contrast;
}

/*
	screen_transition()
	Does a screen transition by changing the contrast setting.
*/
void screen_transition(void)
{
	volatile char *selector	= (char *)0xb4000000;
	volatile char *data	= (char *)0xb4010000;

	*selector = 6;
	*data = 0;

	*selector = 6;
	*data = contrast;
}

//---
//	Screen cast.
//	This method is quite horrible, but still -- it's been working for a
//	hundred frames to make a GIF, nothing more.
//---

int Bfile_CreateFile(const unsigned short *filename, int size);
int Bfile_DeleteFile(const unsigned short *filename);
int Bfile_OpenFile(const unsigned short *filename, int mode);
int Bfile_ReadFile(int handle, void *buffer, int size, int whence);
int Bfile_WriteFile(int handle, const void *data, int size);
int Bfile_CloseFile(int handle);

// This is the header of a 128*64 monochrome bitmap image. Adding a 1024-byte
// gint-style video RAM to this header will construct a valid bmp
// representation of the screen, though vertically inverted.
static const unsigned char bmp_header[0x7e] = {
	0x42, 0x4d, 0x7e, 0x04, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x7e, 0x00, 0x00, 0x00, 0x6c, 0x00,
	0x00, 0x00, 0x80, 0x00, 0x00, 0x00, 0x40, 0x00,
	0x00, 0x00, 0x01, 0x00, 0x01, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x04, 0x00, 0x00, 0x13, 0x0b,
	0x00, 0x00, 0x13, 0x0b, 0x00, 0x00, 0x01, 0x00,
	0x00, 0x00, 0x01, 0x00, 0x00, 0x00, 0x42, 0x47,
	0x52, 0x73, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x02, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00,
	0x00, 0x00, 0xff, 0xff, 0xff, 0x00,
};
// Last error that happened.
static int error_happened = 0;
// Current frame number.
static int frame_no = 0;

/*
	saveTo()
	Saves the given vram to the given file.
*/
void saveTo(unsigned short *filename, void *vram)
{
	Bfile_DeleteFile(filename);
	int x = Bfile_CreateFile(filename, 0x7e + 1024);
	if(x < 0)
	{
		error_happened = x;
		return;
	}

	int fd = Bfile_OpenFile(filename, 0x02);
	if(fd < 0)
	{
		error_happened = x;
		return;
	}

	Bfile_WriteFile(fd, bmp_header, 0x7e);
	Bfile_WriteFile(fd, vram, 1024);
	Bfile_CloseFile(fd);
}

/*
	save_frame()
	Saves the video ram in two files \\fls0\[0-9]{3}[ldm].bmp and
	increments the frame number. Limit is 100 frames.
*/
void save_frame(int gray)
{
	unsigned short filename[] = { '\\', '\\', 'f', 'l', 's', '0', '\\',
		'#', '#', '#', '#', '.', 'b', 'm', 'p', 0x00 };

	// Stop recording the moment an error occur.
	if(error_happened) return;
	if(frame_no >= 100) return;

	filename[7] = frame_no / 100 + '0';
	filename[8] = (frame_no / 10) % 10 + '0';
	filename[9] = frame_no % 10 + '0';
	frame_no++;

	if(gray)
	{
		filename[10] = 'l';
		saveTo(filename, gray_lightVRAM());
		if(error_happened) return;

		filename[10] = 'd';
		saveTo(filename, gray_darkVRAM());
	}
	else
	{
		filename[10] = 'm';
		saveTo(filename, display_getCurrentVRAM());
	}
}
